import unittest

from util import TestFixtureVariant


class TestMax(TestFixtureVariant):
    def fixture_path(self):
        return "../fixtures/max/"


if __name__ == "__main__":
    unittest.main(verbosity=2)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from importlib.resources import files
import os

from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.env.env import Env
from pisek.utils.paths import TaskPath, RawPath, SanitizedPath
from pisek.task_jobs.task_job import TaskJob
from pisek.task_jobs.task_manager import TaskJobManager
from pisek.task_jobs.program import ProgramsJob
from pisek.task_jobs.run_result import RunResult, RunResultKind


class ToolsManager(TaskJobManager):
    """Manager that prepares all tools necessary for task testing."""

    def __init__(self):
        super().__init__("Preparing tools")

    def _get_jobs(self) -> list[Job]:
        self.makedirs(TaskPath.executable_path("."))
        jobs: list[Job] = [
            PrepareMinibox(self._env),
            PrepareTextPreprocessor(self._env),
        ]
        return jobs


class PrepareMinibox(TaskJob):
    """Compiles minibox."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare Minibox", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/minibox.c")
        executable = TaskPath.executable_path("_minibox")
        self._access_file(executable)
        gcc = self._run_subprocess(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-D_GNU_SOURCE",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
                "-Wno-unused-result",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Minibox compilation failed.")


class PrepareTextPreprocessor(TaskJob):
    """Compiles Text Preprocessor."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(env=env, name="Prepare text preprocessor", **kwargs)

    def _run(self):
        source = files("pisek").joinpath("tools/text-preproc.c")
        executable = TaskPath.executable_path("_text-preproc")
        self._access_file(executable)
        gcc = self._run_subprocess(
            [
                "gcc",
                source,
                "-o",
                executable.path,
                "-std=gnu11",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
            ]
        )
        if gcc.returncode != 0:
            raise PipelineItemFailure("Text preprocessor compilation failed.")


class PrepareJudgeLibChecker(TaskJob):
    """Compiles judge from judgelib."""

    def __init__(self, env: Env, checker_name: str, judge: str, **kwargs) -> None:
        self.judge = judge
        super().__init__(env=env, name=f"Prepare {checker_name}", **kwargs)

    def _run(self):
        source_files = ["util.cc", "io.cc", "token.cc", "random.cc", f"{self.judge}.cc"]
        source_dir = files("pisek").joinpath("tools/judgelib")
        sources = [source_dir.joinpath(file) for file in source_files]

        executable = TaskPath.executable_path("_" + self.judge)
        self._access_file(executable)

        gpp = self._run_subprocess(
            [
                "g++",
                *sources,
                "-I",
                source_dir,
                "-o",
                executable.path,
                "-std=gnu++17",
                "-O2",
                "-Wall",
                "-Wextra",
                "-Wno-parentheses",
                "-Wno-sign-compare",
            ]
        )

        if gpp.returncode != 0:
            raise PipelineItemFailure(f"{self.checker_name} compilation failed.")


class PrepareTokenJudge(PrepareJudgeLibChecker):
    """Compiles judge-token from judgelib."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(
            env=env, checker_name="token judge", judge="judge-token", **kwargs
        )


class PrepareShuffleJudge(PrepareJudgeLibChecker):
    """Compiles judge-shuffle from judgelib."""

    def __init__(self, env: Env, **kwargs) -> None:
        super().__init__(
            env=env, checker_name="shuffle judge", judge="judge-shuffle", **kwargs
        )


class SanitizationResultKind(Enum):
    ok = auto()
    changed = auto()
    invalid = auto()
    skipped = auto()


@dataclass(frozen=True)
class SanitizationResult:
    kind: SanitizationResultKind
    msg: str | None = None


class TextPreprocAbstract(ProgramsJob):
    """Abstract job that has method for file sanitization."""

    def _run_text_preproc(
        self, input_: RawPath, output: SanitizedPath
    ) -> SanitizationResult:
        try:
            os.remove(output.path)
        except FileNotFoundError:
            pass

        result = self._run_tool(
            "_text-preproc",
            stdin=input_,
            stdout=output,
            stderr=input_.to_sanitization_log(),
        )
        msg = self._read_file(result.stderr_file).strip()
        if result.returncode == 42:
            if self._files_equal(input_, output):
                return SanitizationResult(SanitizationResultKind.ok, msg)
            else:
                return SanitizationResult(SanitizationResultKind.changed, msg)
        elif result.returncode == 43:
            return SanitizationResult(SanitizationResultKind.invalid, msg)
        else:
            raise RuntimeError(f"Text preprocessor failed on file {input_:p}.")


class SanitizeAbstract(TaskJob):
    def __init__(self, env: Env, input_: TaskPath, output: TaskPath, **kwargs) -> None:
        self.input = input_
        self.output = output
        super().__init__(env=env, **kwargs)

    def _run(self) -> SanitizationResult:
        result = self.prerequisites_results.get("create_source", None)
        if isinstance(result, RunResult) and result.kind != RunResultKind.OK:
            self._copy_file(self.input, self.output)
            return SanitizationResult(SanitizationResultKind.skipped)

        return self._sanitize()

    @abstractmethod
    def _sanitize(self) -> SanitizationResult:
        pass


class Sanitize(SanitizeAbstract, TextPreprocAbstract):
    """Sanitize text file using Text Preprocessor."""

    def __init__(
        self, env: Env, input_: RawPath, output: SanitizedPath, **kwargs
    ) -> None:
        super().__init__(
            env, input_, output, name=f"Sanitize {input_:p} -> {output:p}", **kwargs
        )

    def _sanitize(self):
        return self._run_text_preproc(self.input, self.output)


class IsClean(SanitizeAbstract, TextPreprocAbstract):
    """Check that file is same after sanitizing with Text Preprocessor."""

    def __init__(
        self, env: Env, input_: RawPath, output: SanitizedPath, **kwargs
    ) -> None:
        super().__init__(
            env, input_, output, name=f"Check {input_:p} is clean", **kwargs
        )

    def _sanitize(self):
        result = self._run_text_preproc(self.input, self.output)
        if result.kind != SanitizationResultKind.ok:
            raise PipelineItemFailure(
                f"File {self.input:p} not normalized. (Check encoding, missing newline at the end or '\\r'.)"
            )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pisek.config.task_config import TaskConfig
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    TEST_SPEC,
    verdict_always,
)


def evaluate_verdicts(
    config: TaskConfig, verdicts: list[Verdict], expected: str
) -> tuple[bool, bool, Optional[int]]:
    result = True
    definitive = True
    breaker = None

    for i, mode in enumerate((all, any)):
        oks = list(map(TEST_SPEC[expected][i], verdicts))
        ok = mode(oks)
        result &= ok

        if mode == all:
            definitive &= not ok or TEST_SPEC[expected][i] == verdict_always
            if not ok:
                breaker = oks.index(False)
                break
        elif mode == any:
            definitive &= ok

    return result, definitive, breaker



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import StrEnum, auto
from typing import Optional

from pisek.env.env import Env
from pisek.utils.paths import TESTS_DIR, InputPath, OutputPath


class TestcaseGenerationMode(StrEnum):
    static = auto()
    mixed = auto()
    generated = auto()


@dataclass(frozen=True)
class TestcaseInfo:
    name: str
    repeat: int
    generation_mode: TestcaseGenerationMode
    seeded: bool

    @staticmethod
    def generated(name: str, repeat: int = 1, seeded: bool = True) -> "TestcaseInfo":
        return TestcaseInfo(name, repeat, TestcaseGenerationMode.generated, seeded)

    @staticmethod
    def mixed(name: str) -> "TestcaseInfo":
        return TestcaseInfo(name, 1, TestcaseGenerationMode.mixed, False)

    @staticmethod
    def static(name: str) -> "TestcaseInfo":
        return TestcaseInfo(name, 1, TestcaseGenerationMode.static, False)

    def input_path(
        self, seed: Optional[int] = None, solution: Optional[str] = None
    ) -> InputPath:
        filename = self.name
        if self.seeded:
            assert seed is not None
            filename += f"_{seed:016x}"
        filename += ".in"

        return InputPath(filename, solution=solution)

    def reference_output(
        self, env: Env, seed: Optional[int] = None, solution: Optional[str] = None
    ) -> OutputPath:
        is_static = self.generation_mode == TestcaseGenerationMode.static

        input_path = self.input_path(seed, solution=env.config.primary_solution)
        if is_static:
            path = OutputPath.static(input_path.replace_suffix(".out").name)
        else:
            path = input_path.to_output()

        if solution is not None:
            path = OutputPath(TESTS_DIR, solution, path.name).to_reference_output()

        return path



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fnmatch
from functools import cached_property
import os
from pydantic_core import PydanticCustomError, ErrorDetails
from pydantic import (
    Field,
    computed_field,
    field_validator,
    BeforeValidator,
    ValidationError,
    ValidationInfo,
    model_validator,
)
import re
from typing import Any, Annotated, ClassVar, Mapping, Optional, TypeVar, Union

from pisek.utils.paths import TaskPath
from pisek.utils.text import tab
from pisek.utils.text import eprint, warn
from pisek.utils.colors import ColorSettings
from pisek.env.base_env import BaseEnv
from pisek.config.config_hierarchy import ConfigValue, TaskConfigError, ConfigHierarchy
from pisek.config.config_types import (
    TaskType,
    GenType,
    ValidatorType,
    OutCheck,
    JudgeType,
    ShuffleMode,
    DataFormat,
    TestPoints,
    ProgramRole,
    BuildStrategyName,
    CMSFeedbackLevel,
    CMSScoreMode,
)
from pisek.env.context import init_context
from pisek.task_jobs.solution.solution_result import TEST_SPEC
from pisek.task_jobs.builder.strategies import ALL_STRATEGIES


T = TypeVar("T")
Maybe = Annotated[T | None, BeforeValidator(lambda t: t or None)]


MaybeInt = Annotated[
    Optional[int], BeforeValidator(lambda i: (None if i == "X" else i))
]
ListStr = Annotated[list[str], BeforeValidator(lambda s: s.split())]

TaskPathFromStr = Annotated[TaskPath, BeforeValidator(lambda s: TaskPath(s))]
OptionalTaskPathFromStr = Annotated[
    Optional[TaskPath], BeforeValidator(lambda s: TaskPath(s) if s else None)
]
ListTaskPathFromStr = Annotated[
    list[TaskPath], BeforeValidator(lambda s: [TaskPath(p) for p in s.split()])
]

MISSING_VALIDATION_CONTEXT = "Missing validation context."

ValuesDict = dict[str, Union[str, "ValuesDict", dict[Any, "ValuesDict"]]]
ConfigValuesDict = Mapping[
    str, Union[ConfigValue, "ConfigValuesDict", dict[Any, "ConfigValuesDict"]]
]


def _to_values(config_values_dict: ConfigValuesDict) -> ValuesDict:
    def convert(what: ConfigValue | Mapping) -> str | dict:
        if isinstance(what, ConfigValue):
            return what.value
        else:
            return {key: convert(val) for key, val in what.items()}

    return {key: convert(val) for key, val in config_values_dict.items()}


def _validate_program_name(key: str, value: str) -> str:
    error_key = key.lower().replace(" ", "_")
    for banned_char in "[]":
        if banned_char in value:
            raise PydanticCustomError(
                f"invalid_{error_key}",
                f"{key.capitalize()} must not contain '{banned_char}'",
            )
    if value.startswith("_"):
        raise PydanticCustomError(
            f"invalid_{error_key}",
            f"{key.capitalize()} must not start with '_'",
        )
    return value


class TaskConfig(BaseEnv):
    """Configuration of task loaded from config file."""

    task: "TaskSection"
    tests: "TestsSection"
    test_sections: dict[int, "TestSection"]

    solutions: dict[str, "SolutionSection"]

    solution_time_limit: float = Field(ge=0)  # Needed for visualization

    limits: "LimitsSection"

    cms: "CMSSection"

    checks: "ChecksSection"

    @computed_field  # type: ignore[misc]
    @cached_property
    def total_points(self) -> int:
        return sum(sub.max_points for sub in self.test_sections.values())

    @computed_field  # type: ignore[misc]
    @property
    def tests_count(self) -> int:
        return len(self.test_sections)

    @computed_field  # type: ignore[misc]
    @cached_property
    def input_globs(self) -> list[str]:
        return sum((sub.all_globs for sub in self.test_sections.values()), start=[])

    @computed_field  # type: ignore[misc]
    @property
    def primary_solution(self) -> str:
        if len(self.solutions) == 0:
            raise RuntimeError("No solutions exist.")
        else:
            return [name for name, sol in self.solutions.items() if sol.primary][0]

    def get_solution_by_run(self, run: str) -> Optional[str]:
        sources = (name for name, sol in self.solutions.items() if sol.run.name == run)
        return next(sources, None)

    def __init__(self, **kwargs):
        value = {"test_count": max(kwargs["test_sections"]) + 1}

        with init_context(value):
            super().__init__(**kwargs)

    @staticmethod
    def load_dict(configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, Any] = {}

        args["task"] = TaskSection.load_dict(configs)
        args["tests"] = TestsSection.load_dict(args["task"]["task_type"].value, configs)

        section_names = configs.sections()

        PROGRAMS = [
            (ProgramRole.gen, "in_gen"),
            (ProgramRole.validator, "validator"),
            (ProgramRole.judge, "out_judge"),
        ]
        for t, program in PROGRAMS:
            if args["tests"][program].value:
                args["tests"][program] = RunSection.load_dict(
                    t, args["tests"][program], configs
                )

        # Load tests
        args["test_sections"] = test_sec = {}
        # Sort so tests.keys() returns tests in sorted order
        for section in sorted(section_names, key=lambda cv: cv.value):
            section_name = section.value
            if m := re.fullmatch(r"test(\d{2})", section_name):
                num = m[1]
                test_sec[int(num)] = TestSection.load_dict(
                    ConfigValue(str(int(num)), section.config, section.section, None),
                    configs,
                )

        args["solutions"] = solutions = {}
        for section in section_names:
            if m := re.fullmatch(r"solution_(.+)", section.value):
                solutions[m[1]] = SolutionSection.load_dict(
                    ConfigValue(m[1], section.config, section.section, None), configs
                )

        args["limits"] = LimitsSection.load_dict(configs)
        args["cms"] = CMSSection.load_dict(configs)
        args["checks"] = ChecksSection.load_dict(configs)

        args["solution_time_limit"] = configs.get_from_candidates(
            [("run_solution", "time_limit"), ("run", "time_limit")]
        )

        return args

    @model_validator(mode="after")
    def validate_model(self):
        if (
            self.task.task_type == TaskType.interactive
            and self.tests.out_check != OutCheck.judge
        ):
            raise PydanticCustomError(
                "interactive_must_have_judge",
                "For interactive task 'out_check' must be 'judge'",
                {"task_type": self.task.task_type, "out_check": self.tests.out_check},
            )

        JUDGE_TYPES = {
            TaskType.batch: [
                None,
                JudgeType.opendata_v1,
                JudgeType.opendata_v2,
                JudgeType.cms_batch,
            ],
            TaskType.interactive: [JudgeType.cms_communication],
        }

        if self.tests.judge_type not in JUDGE_TYPES[self.task.task_type]:
            raise PydanticCustomError(
                "task_judge_type_mismatch",
                f"'{self.tests.judge_type}' judge for '{self.task.task_type}' task is not allowed",
                {"task_type": self.task.task_type, "judge_type": self.tests.judge_type},
            )

        primary = [name for name, sol in self.solutions.items() if sol.primary]
        if len(primary) > 1:
            raise PydanticCustomError(
                "multiple_primary_solutions",
                "Multiple primary solutions",
                {"primary_solutions": primary},
            )
        if len(self.solutions) > 0 and len(primary) == 0:
            raise PydanticCustomError(
                "no_primary_solution",
                "No primary solution set",
                {},
            )

        for i in range(len(self.test_sections)):
            if i not in self.test_sections:
                raise PydanticCustomError(
                    "missing_test",
                    f"Missing section [test{i:02}]",
                    {},
                )

        self._compute_predecessors()
        return self

    def _compute_predecessors(self) -> None:
        visited = set()
        computed = set()

        def compute_test(num: int) -> list[int]:
            test = self.test_sections[num]
            if num in computed:
                return test.all_predecessors
            elif num in visited:
                raise PydanticCustomError(
                    "cyclic_predecessor_tests", "Cyclic predecessor tests", {}
                )

            visited.add(num)
            all_predecessors = sum(
                (compute_test(p) for p in test.direct_predecessors),
                start=test.direct_predecessors,
            )

            def normalize_list(l):
                return list(sorted(set(l)))

            test.all_predecessors = normalize_list(all_predecessors)
            test.prev_globs = normalize_list(
                sum(
                    (self.test_sections[p].in_globs for p in test.all_predecessors),
                    start=[],
                )
            )
            test.all_globs = normalize_list(test.prev_globs + test.in_globs)
            computed.add(num)

            return test.all_predecessors

        for i in range(self.tests_count):
            compute_test(i)


class TaskSection(BaseEnv):
    _section: str = "task"
    task_type: TaskType
    score_precision: int = Field(ge=0)

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("task", key) for key in cls.model_fields
        }
        return {"_section": configs.get("task", None), **args}


class TestsSection(BaseEnv):
    _section: str = "tests"

    static_subdir: TaskPathFromStr
    in_gen: Maybe["RunSection"]
    gen_type: Maybe[GenType]

    validator: Maybe["RunSection"]
    validator_type: Maybe[ValidatorType]

    out_check: OutCheck
    out_judge: Maybe["RunSection"]
    judge_type: Maybe[JudgeType]
    judge_needs_in: bool | None
    judge_needs_out: bool | None
    tokens_ignore_newlines: bool | None
    tokens_ignore_case: bool | None
    tokens_float_rel_error: Maybe[float]
    tokens_float_abs_error: Maybe[float]
    shuffle_mode: Maybe[ShuffleMode]
    shuffle_ignore_case: bool | None

    in_format: DataFormat
    out_format: DataFormat

    @staticmethod
    def load_dict(task_type: str, configs: ConfigHierarchy) -> ConfigValuesDict:
        GLOBAL_KEYS = [
            "in_gen",
            "gen_type",
            "validator",
            "validator_type",
            "out_check",
            "in_format",
            "out_format",
            "static_subdir",
        ]
        OUT_CHECK_SPECIFIC_KEYS = [
            ((None, "judge"), "out_judge", ""),
            ((None, "judge"), "judge_type", ""),
            ((TaskType.batch, "judge"), "judge_needs_in", "0"),
            ((TaskType.batch, "judge"), "judge_needs_out", "1"),
            ((None, "tokens"), "tokens_ignore_newlines", "0"),
            ((None, "tokens"), "tokens_ignore_case", "0"),
            ((None, "tokens"), "tokens_float_rel_error", ""),
            ((None, "tokens"), "tokens_float_abs_error", ""),
            ((None, "shuffle"), "shuffle_mode", ""),
            ((None, "shuffle"), "shuffle_ignore_case", "0"),
        ]
        args: dict[str, ConfigValue] = {
            key: configs.get("tests", key) for key in GLOBAL_KEYS
        }

        # Load judge specific keys
        for (task_type_cond, out_check), key, default in OUT_CHECK_SPECIFIC_KEYS:
            if (task_type_cond is None or task_type_cond == task_type) and args[
                "out_check"
            ].value == out_check:
                args[key] = configs.get("tests", key)
            else:
                args[key] = ConfigValue.make_internal(default, "tests", key)

        return {"_section": configs.get("tests", None), **args}

    @model_validator(mode="after")
    def validate_model(self):
        if self.in_gen is not None and self.gen_type is None:
            raise PydanticCustomError(
                "no_gen_type",
                "Missing gen_type",
                {"in_gen": self.in_gen.name, "gen_type": ""},
            )

        if self.validator is not None and self.validator_type is None:
            raise PydanticCustomError(
                "no_validator_type",
                "Missing validator_type",
                {"validator": self.validator.name, "validator_type": ""},
            )

        if (self.tokens_float_abs_error is not None) != (
            self.tokens_float_rel_error is not None
        ):
            raise PydanticCustomError(
                "tokens_errors_must_be_set_together",
                "Both types of floating point error must be set together",
                {
                    "tokens_float_abs_error": self.tokens_float_abs_error,
                    "tokens_float_rel_error": self.tokens_float_rel_error,
                },
            )

        return self


class TestSection(BaseEnv):
    """Configuration of one test (group of testcases)."""

    _section: str
    num: int
    name: str
    points: TestPoints
    in_globs: ListStr
    prev_globs: list[str] = []
    all_globs: list[str] = []
    direct_predecessors: list[int]
    all_predecessors: list[int] = []
    checks_validate: bool
    checks_different_outputs: bool

    @property
    def max_points(self) -> int:
        return 0 if self.points == "unscored" else self.points

    def in_test(self, filename: str) -> bool:
        return any(fnmatch.fnmatch(filename, g) for g in self.all_globs)

    def new_in_test(self, filename: str) -> bool:
        return not any(
            fnmatch.fnmatch(filename, g) for g in self.prev_globs
        ) and self.in_test(filename)

    @staticmethod
    def load_dict(number: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "name",
            "points",
            "in_globs",
            "predecessors",
            "checks.validate",
            "checks.different_outputs",
        ]
        args: dict[str, Any] = {
            key.replace(".", "_"): configs.get_from_candidates(
                [(number.section, key), ("tests", key)]
            )
            for key in KEYS
        }
        args["direct_predecessors"] = args.pop("predecessors")

        return {"_section": configs.get(number.section, None), "num": number, **args}

    @field_validator("in_globs", mode="after")
    @classmethod
    def validate_globs(cls, value: list[str], info: ValidationInfo) -> list[str]:
        globs = []
        for glob in value:
            if glob == "@ith":
                glob = f"{info.data['num']:02}*.in"
            if not glob.endswith(".in"):
                raise PydanticCustomError(
                    "in_globs_end_in", "In_globs must end with '*.in'"
                )
            globs.append(glob)

        return globs

    @field_validator("direct_predecessors", mode="before")
    @classmethod
    def expand_predecessors(cls, value: str, info: ValidationInfo) -> list[str]:
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        test_cnt = info.context.get("test_count")
        number = info.data["num"]

        predecessors = []
        for pred in value.split():
            if pred == "@previous":
                if number <= 1:
                    continue
                predecessors.append(number - 1)
            else:
                try:
                    num = int(pred)
                except ValueError:
                    raise PydanticCustomError(
                        "predecessors_must_be_int", "Predecessors must be int"
                    )
                if not 0 <= num < test_cnt:
                    raise PydanticCustomError(
                        "predecessors_must_be_in_range",
                        f"Predecessors must be in range 0, {test_cnt-1}",
                    )
                predecessors.append(num)

        return list(sorted(set(predecessors)))

    @model_validator(mode="after")
    def validate_model(self):
        if self.name == "@auto":
            self.name = f"Test {self.num}"

        return self


class SolutionSection(BaseEnv):
    """Configuration of one solution."""

    _section: str
    name: str
    primary: bool
    run: "RunSection"
    points: MaybeInt
    points_min: MaybeInt
    points_max: MaybeInt
    tests: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def load_dict(cls, name: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        KEYS = [
            "primary",
            "run",
            "points",
            "points_min",
            "points_max",
            "tests",
        ]
        args = {
            key: configs.get_from_candidates([(name.section, key), ("solutions", key)])
            for key in KEYS
        }

        if args["run"].value == "@auto":
            args["run"].value = name.value

        sol_type = (
            ProgramRole.primary_solution
            if args["primary"].value in ()
            else ProgramRole.secondary_solution
        )

        return {
            "_section": configs.get(name.section, None),
            "name": name,
            **args,
            "run": RunSection.load_dict(sol_type, args["run"], configs),
        }

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, value: str) -> str:
        return _validate_program_name("solution name", value)

    @field_validator("tests", mode="after")
    def validate_tests(cls, value, info: ValidationInfo):
        if info.context is None:
            raise RuntimeError(MISSING_VALIDATION_CONTEXT)
        test_cnt = info.context.get("test_count")
        primary = info.data.get("primary")
        if value == "@auto":
            value = ("1" if primary else "X") * test_cnt
        elif value == "@all":
            value = "1" * test_cnt
        elif value == "@any":
            value = "X" * test_cnt

        if len(value) != test_cnt:
            raise PydanticCustomError(
                "tests_str_invalid_len",
                f"There are {test_cnt} tests but test string has {len(value)} characters",
            )

        for char in value:
            if char not in TEST_SPEC:
                raise PydanticCustomError(
                    "tests_str_invalid_char",
                    f"Not allowed char in test string: {char}. Recognized are {''.join(TEST_SPEC.keys())}",
                )

        if primary and value != "1" * test_cnt:
            raise PydanticCustomError(
                "primary_sol_must_succeed",
                f"Primary solution must have: tests={'1'*test_cnt}",
            )

        return value

    @model_validator(mode="after")
    def validate_model(self):
        for points_limit in ["points_min", "points_max"]:
            if self.points is not None and getattr(self, points_limit) is not None:
                raise PydanticCustomError(
                    "points_double_set",
                    f"Both 'points' and '{points_limit}' are set at once",
                    {"points": self.points, points_limit: getattr(self, points_limit)},
                )

        return self


def get_run_defaults(program_role: ProgramRole, program_name: str) -> list[str]:
    if program_role.is_solution():
        return [
            f"run_solution:{program_name}",
            f"run_{program_role}",
            f"run_solution",
            f"run",
        ]
    else:
        return [
            f"run_{program_role}:{program_name}",
            f"run_{program_role}",
            f"run",
        ]


class RunSection(BaseEnv):
    """Configuration of running an program"""

    _section: str

    program_role: ProgramRole
    name: str
    subdir: str
    build: "BuildSection"
    exec: TaskPathFromStr
    time_limit: float = Field(ge=0)  # [seconds]
    clock_mul: float = Field(ge=0)  # [1]
    clock_min: float = Field(ge=0)  # [seconds]
    mem_limit: int = Field(ge=0)  # [KB]
    process_limit: int = Field(ge=0)  # [1]
    # limit=0 means unlimited
    args: ListStr
    env: dict[str, str]

    @property
    def executable(self) -> TaskPath:
        return TaskPath.executable_path(self.build.program_name, self.exec.path)

    def clock_limit(self, override_time_limit: Optional[float] = None) -> float:
        tl = override_time_limit if override_time_limit is not None else self.time_limit
        if tl == 0:
            return 0
        return max(tl * self.clock_mul, self.clock_min)

    @classmethod
    def load_dict(
        cls, program_role: ProgramRole, name: ConfigValue, configs: ConfigHierarchy
    ) -> ConfigValuesDict:
        default_sections = get_run_defaults(program_role, name.value)

        section_name = configs.get_from_candidates(
            [(section, None) for section in default_sections]
        )
        args = {
            key: configs.get_from_candidates(
                [(section, key) for section in default_sections]
            )
            for key in cls.model_fields
            if key not in ("name", "program_role", "env")
        }
        if args["build"].value == "@auto":
            args["build"].value = (
                f"{program_role.build_name}:{os.path.join(args['subdir'].value, name.value)}"
            )

        envs = configs.get_regex_from_candidates(
            (section, "env_(.*)") for section in default_sections
        )
        envs = {key.removeprefix("env_"): val for key, val in envs.items()}

        return {
            "_section": section_name,
            "program_role": ConfigValue.make_internal(
                program_role.name, "run", "program_role"
            ),
            "name": name,
            **args,
            "build": BuildSection.load_dict(args["build"], configs),
            "env": envs,
        }

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, value: str) -> str:
        return _validate_program_name("program name", value)


class BuildSection(BaseEnv):
    program_names: ClassVar[dict[str, str]] = {}

    _section: str
    section_name: str
    build_type: str
    program_name: str

    sources: ListTaskPathFromStr
    comp_args: ListStr
    extras: ListTaskPathFromStr
    strategy: BuildStrategyName
    entrypoint: str

    headers_c: ListTaskPathFromStr
    extra_sources_c: ListTaskPathFromStr
    headers_cpp: ListTaskPathFromStr
    extra_sources_cpp: ListTaskPathFromStr
    extra_sources_py: ListTaskPathFromStr
    extra_sources_java: ListTaskPathFromStr

    @classmethod
    def load_dict(cls, name: ConfigValue, configs: ConfigHierarchy) -> ConfigValuesDict:
        program = name
        program_role = ConfigValue("", name.config, name.section, name.key)
        default_sections = [f"build:{program.value}", "build"]
        for pt in ProgramRole:
            prefix = f"{pt.build_name}:"
            if name.value.startswith(prefix):
                program_role, program = name.split(":")
                default_sections = [
                    f"build_{pt.build_name}:{program.value}",
                    f"build_{pt.build_name}",
                    f"build",
                ]
                break

        if (
            program.value in cls.program_names
            and cls.program_names[program.value] != default_sections[0]
        ):
            raise TaskConfigError(
                "Colliding suffixes of build sections not allowed: "
                f"[{default_sections[0]}] and [{cls.program_names[program.value]}]."
            )
        cls.program_names[program.value] = default_sections[0]

        section_name = configs.get_from_candidates(
            [(section, None) for section in default_sections]
        )
        args = {
            key: configs.get_from_candidates(
                [(section, key) for section in default_sections]
            )
            for key in cls.model_fields
            if key not in ("section_name", "build_type", "program_name")
        }

        return {
            "_section": section_name,
            "section_name": ConfigValue.make_internal(
                default_sections[0], default_sections[0], None
            ),
            "build_type": program_role,
            "program_name": program,
            **args,
        }

    @field_validator("program_name", mode="after")
    @classmethod
    def validate_program_name(cls, value: str) -> str:
        return _validate_program_name("program name", value)

    @field_validator("sources", mode="before")
    @classmethod
    def convert_sources(cls, value: str, info: ValidationInfo) -> str:
        if value == "@auto":
            return str(info.data.get("program_name"))
        else:
            return value


class LimitsSection(BaseEnv):
    """Configuration of input and output size limits."""

    _section: str = "limits"

    input_max_size: int
    output_max_size: int

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("limits", key) for key in cls.model_fields
        }
        return {"_section": configs.get("limits", None), **args}


class CMSSection(BaseEnv):
    _section: str = "cms"

    name: Maybe[str]
    title: str
    submission_format: ListStr

    time_limit: float = Field(gt=0)  # [seconds]
    mem_limit: int = Field(gt=0)  # [KB]

    max_submissions: MaybeInt = Field(gt=0)
    min_submission_interval: int = Field(ge=0)  # [seconds]

    score_mode: CMSScoreMode
    feedback_level: CMSFeedbackLevel

    stubs: ListTaskPathFromStr
    headers: ListTaskPathFromStr

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args = {key: configs.get("cms", key) for key in cls.model_fields}

        def get_strategy_union(key: str) -> str:
            all_items = {
                configs.get_from_candidates(
                    [
                        ("build_solution", getattr(strat, key)),
                        ("build", getattr(strat, key)),
                    ]
                ).value
                for strat in ALL_STRATEGIES.values()
                if getattr(strat, key) is not None
            }
            return " ".join(sorted(all_items))

        if args["stubs"].value == "@auto":
            args["stubs"].value = get_strategy_union("extra_sources")
        if args["headers"].value == "@auto":
            args["headers"].value = get_strategy_union("extra_nonsources")

        return {"_section": configs.get("cms", None), **args}

    @field_validator("title", mode="before")
    @classmethod
    def convert_title(cls, value: str, info: ValidationInfo) -> str:
        if value == "@name":
            name = info.data.get("name")
            name = "unnamed-task" if name is None else name

            return name
        else:
            return value

    @field_validator("submission_format", mode="after")
    @classmethod
    def convert_format(cls, value: list[str], info: ValidationInfo) -> list[str]:
        name = info.data.get("name")
        name = "unnamed-task" if name is None else name

        return [
            (CMSSection.get_default_file_name(name) if n == "@name" else n)
            for n in value
        ]

    @classmethod
    def get_default_file_name(cls, name: str):
        name = re.sub(r"[^a-zA-Z0-9]+", "_", name)
        return f"{name}.%l"


class ChecksSection(BaseEnv):
    """Configuration of checks for pisek to run."""

    _section: str = "checks"

    solution_for_each_test: bool
    no_unused_inputs: bool
    all_inputs_in_last_test: bool
    generator_respects_seed: bool
    one_input_in_each_nonsample_test: bool
    fuzzing_thoroughness: int

    @classmethod
    def load_dict(cls, configs: ConfigHierarchy) -> ConfigValuesDict:
        args: dict[str, ConfigValue] = {
            key: configs.get("checks", key) for key in cls.model_fields
        }
        return {"_section": configs.get("checks", None), **args}


def _format_message(err: ErrorDetails) -> str:
    inp = err["input"]
    ctx = err["ctx"] if "ctx" in err else None
    if isinstance(inp, dict) and ctx is not None:
        if ctx == {}:
            return f"{err['msg']}."
        return f"{err['msg']}:\n" + tab(
            "\n".join(f"{key}={val}" for key, val in ctx.items())
        )
    return f"{err['msg']}: '{inp}'"


def _convert_errors(e: ValidationError, config_values: ConfigValuesDict) -> list[str]:
    error_msgs: list[str] = []
    for error in e.errors():
        value: Any = config_values
        for loc in error["loc"]:
            value = value[loc]

        if not isinstance(value, ConfigValue):
            location = (
                value["_section"].location() if "_section" in value else "global config"
            )
        else:
            location = value.location()

        error_msgs.append(f"In {location}:\n" + tab(_format_message(error)))
    return error_msgs


def load_config(
    path: str,
    pisek_directory: Optional[str],
    config_filename: str,
    strict: bool = False,
    suppress_warnings: bool = False,
) -> Optional[TaskConfig]:
    """Loads config from given path."""
    try:
        config_hierarchy = ConfigHierarchy(
            path, not suppress_warnings, pisek_directory, config_filename
        )
        config_values = TaskConfig.load_dict(config_hierarchy)
        config = TaskConfig(**_to_values(config_values))
        config_hierarchy.check_all()
        if config_hierarchy.check_todos() and not suppress_warnings:
            warn("Unsolved TODOs in config.", TaskConfigError, strict)
        return config
    except TaskConfigError as err:
        eprint(ColorSettings.colored(str(err), "red"))
    except ValidationError as err:
        eprint(
            ColorSettings.colored(
                "Invalid config:\n\n"
                + "\n\n".join(_convert_errors(err, config_values)),
                "red",
            )
        )
    return None



"""
Tests the command-line interface.
"""

import os

import unittest
from io import StringIO
from unittest import mock

from util import TestFixture

from pisek.__main__ import main


class TestCLI(TestFixture):
    def fixture_path(self):
        return "../fixtures/sum_cms/"

    def args(self):
        return [["test", "--time-limit", "0.2"]]

    def runTest(self):
        if not self.fixture_path():
            return

        self.log_files()

        with mock.patch("sys.stdout", new=StringIO()) as std_out:
            with mock.patch("sys.stderr", new=StringIO()) as std_err:
                for args_i in self.args():
                    result = main(args_i)

                    self.assertFalse(
                        result,
                        f"Command failed: {' '.join(args_i)}",
                    )

        self.check_files()


class TestCLITestSolution(TestCLI):
    def args(self):
        return [["test", "solutions", "solve"]]


class TestCLITestGenerator(TestCLI):
    def args(self):
        return [["test", "generator"]]


class TestCLIClean(TestCLI):
    def args(self):
        return [["clean"]]


class TestCLITestingLog(TestCLI):
    def args(self):
        return [["test", "--testing-log"]]

    def created_files(self):
        return ["testing_log.json"]


class TestCLIVisualize(TestCLI):
    def args(self):
        return [["test", "--testing-log"], ["visualize"]]

    def created_files(self):
        return ["testing_log.json"]


if __name__ == "__main__":
    unittest.main(verbosity=2)



from configparser import ConfigParser
from functools import partial
from importlib.resources import files
import os
import shutil
from typing import Callable, Sequence

from pisek.utils.text import eprint
from pisek.utils.colors import ColorSettings
from pisek.utils.user_input import input_string, input_choice
from pisek.config.config_types import (
    TaskType,
    GenType,
    ValidatorType,
    OutCheck,
    JudgeType,
)
from pisek.config.config_hierarchy import DEFAULT_CONFIG_FILENAME
from pisek.task_jobs.program import ProgramRole

EXAMPLE_TASKS_DIR = str(files("pisek").joinpath("../examples"))
EXAMPLE_TASKS = ["cms-batch"]

OFFERED_GEN_TYPES = [GenType.pisek_v1, GenType.opendata_v1, GenType.cms_old]
RECOMMENDED_GEN_TYPES = [GenType.pisek_v1]

OFFERED_VALIDATOR_TYPES = [ValidatorType.simple_42]
RECOMMENDED_VALIDATOR_TYPES: list[ValidatorType] = (
    []
)  # All offered validator types are reasonable, no need to recommend anything

OFFERED_JUDGE_TYPES = {
    TaskType.batch: [JudgeType.cms_batch, JudgeType.opendata_v2],
    TaskType.interactive: [JudgeType.cms_communication],
}
RECOMMENDED_JUDGE_TYPES: dict[TaskType, list[JudgeType]] = (
    {  # All offered judge types are reasonable, no need to recommend anything
        TaskType.batch: [],
        TaskType.interactive: [],
    }
)

SOLUTION_SUBDIR = "solutions"


def remove_suffix(path: str) -> str:
    return os.path.splitext(path)[0]


def touch(path: str) -> None:
    open(path, "a").close()


def recommended() -> str:
    return " (" + ColorSettings.colored("recommended", "green") + ")"


def invalid_config_name(config_filename: str) -> int:
    eprint(
        ColorSettings.colored(
            f"Config filename '{config_filename}' already in use.",
            "red",
        )
    )
    return 2


def input_program(
    p_role: ProgramRole, types: Sequence[str], recommended_types: Sequence[str]
) -> tuple[str, str]:

    choices = [
        (t, t + (recommended() if t in recommended_types else "")) for t in types
    ]
    type_ = input_choice(f"Choose {p_role}_type:", choices)
    filename = input_string(f"Enter {p_role} filename: ")
    print()
    return type_, filename


def init_task(config_filename: str) -> int:
    if os.listdir():
        eprint(ColorSettings.colored("Current directory is not empty", "red"))
        return 1

    example_tasks: list[tuple[Callable[[str], int], str]] = [
        (
            partial(from_template, os.path.join(EXAMPLE_TASKS_DIR, task)),
            task + " example task",
        )
        for task in EXAMPLE_TASKS
    ]
    create = input_choice(
        "Create a task", [(from_scratch, "From scratch")] + example_tasks
    )
    return create(config_filename)


def from_scratch(config_filename: str) -> int:
    config = ConfigParser(interpolation=None)
    config.add_section("task")
    config.add_section("tests")
    config["task"]["version"] = "v3"

    # --- task_type ---
    print()
    task_type = input_choice("Choose task_type:", [(t, t) for t in TaskType])
    config["task"]["task_type"] = task_type
    print()

    # --- inputs ---
    gen_type, gen_name = input_program(
        ProgramRole.gen, OFFERED_GEN_TYPES, RECOMMENDED_GEN_TYPES
    )
    touch(gen_name)
    config["tests"]["in_gen"] = remove_suffix(gen_name)
    config["tests"]["gen_type"] = gen_type

    touch("sample.in")
    touch("sample.out")

    # --- validator ---
    val_type, val_name = input_program(
        ProgramRole.validator, OFFERED_VALIDATOR_TYPES, RECOMMENDED_VALIDATOR_TYPES
    )
    touch(val_name)
    config["tests"]["validator"] = remove_suffix(val_name)
    config["tests"]["validator_type"] = val_type

    # --- out_check ---
    if task_type == TaskType.batch:
        out_check = input_choice(
            "How to check outputs?",
            [
                (OutCheck.tokens, "tokens - output is unique"),
                (OutCheck.shuffle, "shuffle - output is unique up to order"),
                (OutCheck.judge, "judge - output is not unique"),
            ],
        )
        print()
    else:
        out_check = OutCheck.judge

    config["tests"]["out_check"] = out_check

    if out_check == OutCheck.judge:
        judge_type, judge_name = input_program(
            ProgramRole.judge,
            OFFERED_JUDGE_TYPES[task_type],
            RECOMMENDED_JUDGE_TYPES[task_type],
        )
        touch(judge_name)
        config["tests"]["out_judge"] = remove_suffix(judge_name)
        config["tests"]["judge_type"] = judge_type

    # --- solutions ---
    os.makedirs(SOLUTION_SUBDIR)
    config.add_section("run_solution")
    config["run_solution"]["time_limit"] = "1"
    config["run_solution"]["subdir"] = "solutions"

    sol = input_string("Enter primary solution filename: ")
    touch(os.path.join(SOLUTION_SUBDIR, sol))
    sol_sec = f"solution_{remove_suffix(sol)}"
    config.add_section(sol_sec)
    config[sol_sec]["primary"] = "yes"

    try:
        with open(config_filename, "x") as f:
            config.write(f, space_around_delimiters=False)
    except FileExistsError:
        return invalid_config_name(config_filename)

    print()
    print("For more information visit our docs: https://piskoviste.github.io/pisek/")

    return 0


def from_template(path: str, config_filename: str) -> int:
    if os.path.exists(os.path.join(path, config_filename)):
        return invalid_config_name(config_filename)

    for item in os.listdir(path):
        s = os.path.join(path, item)
        if os.path.isdir(s):
            shutil.copytree(s, item)
        else:
            shutil.copy(s, item)

    shutil.move(DEFAULT_CONFIG_FILENAME, config_filename)

    return 0



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Antonín Maloň <git@tonyl.eu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from decimal import Decimal
import json
from math import ceil, inf
import os
from typing import Any, Optional

from pisek.utils.text import pad, tab, eprint
from pisek.utils.colors import ColorSettings
from pisek.utils.terminal import terminal_width
from pisek.config.config_types import TestPoints
from pisek.config.task_config import load_config, TaskConfig
from pisek.config.select_solutions import expand_solutions, UnknownSolutions
from pisek.task_jobs.solution.solution_result import Verdict
from pisek.task_jobs.solution.verdicts_eval import evaluate_verdicts


class MissingSolution(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(f"Missing solution {name} in testing log.")


@dataclass
class LoggedResult:
    verdict: Verdict
    relative_points: Optional[Decimal]
    absolute_points: Optional[Decimal]
    time: float
    test: str
    original_verdict: Verdict

    def points(self, test_points: TestPoints) -> Decimal:
        if test_points == "unscored":
            return Decimal("0")
        if self.relative_points is not None:
            return self.relative_points * test_points
        elif self.absolute_points is not None:
            return self.absolute_points
        else:
            raise ValueError("Relative and absolute points unset")

    def to_str(
        self,
        limit: float,
        segments: int,
        test_pad_length: int = 15,
    ) -> str:
        def get_bar() -> str:
            percentage = self.time / limit
            full = max(ceil(segments * percentage), 1)

            bounded = min(full, segments)
            overflown = max(0, full - segments)

            overflown_max_length = terminal_width - segments - test_pad_length - 40
            if cut := (overflown > overflown_max_length):
                overflown = overflown_max_length

            if abs(percentage - 1) < 0.1:
                color = "red"
            elif abs(percentage - 1) < 0.5:
                color = "yellow"
            else:
                color = "green"

            full_bar = ColorSettings.colored("━", color)

            return (
                f"[{full_bar*bounded}{('━' if ColorSettings.colors_on else ' ')*(segments-bounded)}|"
                f"{full_bar*overflown}{' '*(overflown_max_length - overflown)}{'⋯⋯' if cut else '  '}"
            )

        return (
            f"{pad(self.test, test_pad_length)}  "
            f"{self.original_verdict.mark()}->{self.verdict.mark()}  "
            f"{get_bar()}    {self.time:.2f} / {limit:.2f}s"
        )


def limit_result(result: LoggedResult, limit: float) -> LoggedResult:
    if result.time <= limit and result.verdict == Verdict.timeout:
        new_verdict = result.original_verdict
        if new_verdict == Verdict.timeout:
            new_verdict = Verdict.ok

        return LoggedResult(
            new_verdict,
            Decimal(1),
            None,
            result.time,
            result.test,
            result.original_verdict,
        )
    if result.time > limit:
        return LoggedResult(
            Verdict.timeout,
            Decimal(0),
            None,
            result.time,
            result.test,
            result.original_verdict,
        )
    return result


class SolutionResults:
    def __init__(
        self, name: str, config: TaskConfig, results: list[LoggedResult]
    ) -> None:
        self.name = name
        self._solution = config.solutions[name]
        self._config = config
        self._results = results

        self._results.sort(
            key=lambda r: (self._get_test(r), r.verdict.value, r.time, r.test)
        )

    def _get_test(self, result: LoggedResult) -> int:
        return min(
            i
            for i, sub in self._config.test_sections.items()
            if sub.new_in_test(result.test)
        )

    def _evaluate_results(
        self, results: list[LoggedResult], test_num: int
    ) -> tuple[bool, bool, Optional[LoggedResult]]:
        ok, definitive, breaker = evaluate_verdicts(
            self._config,
            list(map(lambda r: r.verdict, results)),
            self._solution.tests[test_num],
        )
        return ok, definitive, results[breaker] if breaker is not None else None

    @staticmethod
    def from_log(
        name: str, config: TaskConfig, testing_log, limit: float
    ) -> "SolutionResults":
        if name not in testing_log["solutions"]:
            raise MissingSolution(name)

        results = []
        for test_name, result in testing_log["solutions"][name]["results"].items():

            def load_decimal(result: dict[str, Any], key: str) -> Optional[Decimal]:
                return Decimal(result[key]) if key in result else None

            results.append(
                limit_result(
                    LoggedResult(
                        Verdict[result["result"]],
                        load_decimal(result, "relative_points"),
                        load_decimal(result, "absolute_points"),
                        result["time"],
                        test_name,
                        Verdict[result["result"]],
                    ),
                    limit,
                )
            )

        return SolutionResults(name, config, results)

    def get_all(self) -> list[LoggedResult]:
        return self._results

    def get_by_test(self) -> list[list[LoggedResult]]:
        by_test: list[list[LoggedResult]] = [
            [] for _ in range(self._config.tests_count)
        ]
        for res in self._results:
            for num, sub in self._config.test_sections.items():
                if sub.in_test(res.test):
                    by_test[num].append(res)

        return by_test

    def check_test(self, num: int) -> Optional[str]:
        results = self.get_by_test()

        expected = self._solution.tests[num]
        ok, _, breaker = self._evaluate_results(results[num], num)
        if not ok:
            failed_test = f" ({breaker.test})" if breaker else ""
            return f"{expected}{failed_test}"
        return None

    def check_points(self) -> Optional[str]:
        achieved = Decimal(0)
        results = self.get_by_test()
        for num, sub in self._config.test_sections.items():
            achieved += min(map(lambda r: r.points(sub.points), results[num]))

        points = self._solution.points
        points_min = self._solution.points_min
        points_max = self._solution.points_max

        if points is not None and points != achieved:
            return f"{points}p"
        if points_min is not None and not achieved >= points_min:
            return f"at least {points_min}p"
        if points_max is not None and not achieved <= points_max:
            return f"at most {points_max}p"

        return None

    def check_all(self) -> list[str]:
        fails = []
        for num in self._config.test_sections:
            if (err := self.check_test(num)) is not None:
                fails.append(f"{err} on {self._config.test_sections[num].name}")
        if (err := self.check_points()) is not None:
            fails.append(f"get {err}")

        return fails

    def get_time_limit_range(self, num: int) -> tuple[float, float]:
        results = self.get_by_test()[num]
        times = [0] + list(map(lambda r: r.time, results))
        times.sort()
        min_possible = len(times) - 1
        max_possible = 0
        for i, time in enumerate(times):
            ok, _, _ = self._evaluate_results(
                list(map(lambda r: limit_result(r, time), results)),
                num,
            )
            if ok:
                min_possible = min(i, min_possible)
                max_possible = max(i, max_possible)

        times.append(inf)
        return (times[min_possible], times[max_possible + 1])


def show_all(results: list[LoggedResult]) -> list[LoggedResult]:
    return results


def show_slowest(results: list[LoggedResult]) -> list[LoggedResult]:
    return [max(results, key=lambda r: (r.verdict.value, r.time))]


def visualize(
    path: str,
    filter: str,
    bundle: bool,
    solutions: Optional[list[str]],
    limit: Optional[float],
    filename: str,
    segments: Optional[int],
    pisek_dir: Optional[str],
    config_filename: str,
    **_,
) -> int:
    config = load_config(path, pisek_dir, config_filename)
    if config is None:
        return 2

    log_path = os.path.join(path, filename)
    try:
        with open(log_path) as log_file:
            testing_log = json.load(log_file)
    except FileNotFoundError:
        eprint(
            ColorSettings.colored(
                f"File {log_path} not found. Test with --testing-log to create a log.",
                "red",
            )
        )
        return 2

    if testing_log["source"] == "cms":
        limit_default = config.cms.time_limit
    else:
        limit_default = config.solution_time_limit
    time_limit = limit_default if limit is None else limit

    filter_fn = show_all if filter == "all" else show_slowest

    segment_cnt = terminal_width // 8 if segments is None else segments

    try:
        expanded_solutions = expand_solutions(config, solutions)
    except UnknownSolutions as err:
        eprint(ColorSettings.colored(str(err), "red"))
        return 2

    results: dict[str, SolutionResults] = {}
    max_test_length = 0
    for sol in expanded_solutions:
        try:
            results[sol] = SolutionResults.from_log(
                sol, config, testing_log, time_limit
            )
            max_test_length = max(
                max_test_length, max(map(lambda r: len(r.test), results[sol].get_all()))
            )
        except MissingSolution as err:
            eprint(ColorSettings.colored(str(err), "yellow"))

    wrong_solutions = {}
    for sol, sol_res in results.items():
        if not bundle:
            sol_err = sol_res.check_points()
            err_msg = ""
            if sol_err is not None:
                err_msg = ColorSettings.colored(f" should get {sol_err}", "red")
                wrong_solutions[sol] = True
            print(f"{sol}{err_msg}")

            for num, group_res in enumerate(sol_res.get_by_test()):
                test_err = sol_res.check_test(num)
                err_msg = ""
                if test_err is not None:
                    err_msg = ColorSettings.colored(f" should result {test_err}", "red")
                    wrong_solutions[sol] = True

                print(tab(f"{config.test_sections[num].name}{err_msg}"))
                for res in filter_fn(group_res):
                    print(
                        tab(tab(res.to_str(time_limit, segment_cnt, max_test_length)))
                    )

        else:
            sol_errs = sol_res.check_all()
            err_msg = ""
            if sol_errs:
                err_msg = ColorSettings.colored(f" should {', '.join(sol_errs)}", "red")
                wrong_solutions[sol] = True
            print(f"{sol}{err_msg}")

            for res in filter_fn(sol_res.get_all()):
                print(tab(tab(res.to_str(time_limit, segment_cnt, max_test_length))))

    print()
    if wrong_solutions:
        print(
            ColorSettings.colored(
                f"Solutions {', '.join(wrong_solutions.keys())} should result differently.",
                "red",
            )
        )

    min_possible = 0.0
    max_possible = inf
    for sol, sol_res in results.items():
        for num in config.test_sections:
            a, b = sol_res.get_time_limit_range(num)
            min_possible = max(a, min_possible)
            max_possible = min(b, max_possible)

    if min_possible <= max_possible:
        limit_msg = f"Valid time limit between {min_possible:.2f}, {max_possible:.2f}."
    else:
        limit_msg = "No valid time limit found."
    print(ColorSettings.colored(limit_msg, "cyan"))

    return 0



import os
from typing import Optional

from pisek.config.update_config import update_config
from pisek.config.config_hierarchy import new_config_parser
from pisek.config.task_config import load_config


def update_and_replace_config(
    task_path: str,
    pisek_directory: Optional[str],
    config_filename: str,
) -> bool:
    if (
        load_config(task_path, pisek_directory, config_filename, suppress_warnings=True)
        is None
    ):
        return False  # Raise errors if config is invalid

    config_path = os.path.join(task_path, config_filename)
    config = new_config_parser()
    config.read(config_path)
    update_config(config, task_path=task_path, infos=False)
    with open(config_path, "w") as f:
        config.write(f, space_around_delimiters=False)

    return True
